﻿//------------------------------------------------------------------------------
//	ログ･スケールのグラフを描くためのクラスの拡張
//
//		このクラスについては，下記の書籍を参照のこと
//		三上直樹：“理系のための ｢C#｣ プログラミング”，第８章，工学社，2012年
//
//		作成者：三上直樹，2017/12/01 作成，©三上直樹 2017
//------------------------------------------------------------------------------

using System;
using System.Windows.Forms;
using System.Drawing;
using System.Diagnostics;
using System.Drawing.Drawing2D;

using static System.Math;

namespace Mikami.GraphicsLib
{
	public class LogScale
	{
		// 縦方向描画のモードに対する列挙型の定義：リニア，ログ のいずれか
		private enum Mode { Log, Linear }

		private readonly Values x_;		// x 軸用
		private readonly Values y_;		// y 軸用
		private readonly Graphics gr_;	// グラフィックス用
		private readonly Mode mode_;	// 描画のモード用
		private readonly double eps_;	// ｢Y()｣ で値が 0 の場合に置換える値

		// 描画に必要な定数を格納するための構造体
		public struct Values
		{
			public readonly float org, min, max, dec;
			public Values(double u0, double uMin, double uMax, double uDec)
			{
				org = (float)u0;		// 座標軸の目盛の最小値に対応する位置
				min = (float)uMin;		// 座標軸の目盛の最小値
				max = (float)uMax;		// 座標軸の目盛の最大値
				dec = (float)uDec;		// 10 倍に対応する長さ
			}
		}

		// コンストラクタ
		public LogScale(Graphics gr, Values u, Values? v = null)
		{
			gr_ = gr;
			x_ = u;
			y_ = v ?? new Values();
			mode_ = (v == null) ? Mode.Linear : Mode.Log;
			if (v != null) eps_ = 1.0e-4*y_.min;
		}

		// x 座標値の取得
		public float X(double x)
		{
			return x_.org + x_.dec*(float)Log10(x/x_.min);
		}

		// x 座標値から周波数への変換
		public float ToFrq(double x)
		{
			return x_.min*(float)Pow(10, (x-x_.org)/x_.dec);
		}

		// y 座標値の取得
		public float Y(double y)
		{
			Test(Mode.Linear, "Linear モードで Y( ) は使用できません");
			if (y < eps_) y = eps_;
			return y_.org - y_.dec*(float)Log10(y/y_.min);
		}

		// x 軸の数値に対する目盛線の描画
		public void DrawAxisX(Pen aPen, float y0, float h1, float h2 = 0)
		{
			if (h2 == 0) h2 = h1;
			DrawScale(x_, x => gr_.DrawLine(aPen, X(x), y0, X(x),
											y0-(PowersOf10(x) ? h1 : h2)));
		}

		// y 軸の数値に対する目盛線の描画
		public void DrawAxisY(Pen aPen, float x0, float w1, float w2 = 0)
		{
			if (w2 == 0) w2 = w1;
			DrawScale(y_, y => gr_.DrawLine(aPen, x0, Y(y),
											x0+(PowersOf10(y) ? w1 : w2), Y(y)));
		}

		// グラフの描画，Graphics.DrawLines() メソッドを使用
		public void DrawLines(Func<double, double> y, int nPoints, Pen aPen,
							  float y0 = 0, float size = 0,
							  SmoothingMode mode = SmoothingMode.AntiAlias)
		{
			DrawGraph(y, nPoints, aPen, gr_.DrawLines, y0, size, mode);
		}

		// グラフの描画，Graphics.DrawCurve() メソッドを使用
		public void DrawCurve(Func<double, double> y, int nPoints, Pen aPen,
							  float y0 = 0, float size = 0,
							  SmoothingMode mode = SmoothingMode.AntiAlias)
		{
			DrawGraph(y, nPoints, aPen, gr_.DrawCurve, y0, size, mode);
		}

		// DrawLines(), DrawCurve() で使用
		private void DrawGraph(Func<double, double> y, int nPoints, Pen aPen,
							   Action<Pen, PointF[]> Draw,
							   float y0, float size, SmoothingMode mode)
		{
			// 隣同士の描画点に対応する x 軸方向の値の比
			double ratio = Pow(x_.max/x_.min, 1.0/(nPoints - 1));

			Func<double, float> y_x = x =>
					(mode_ == Mode.Linear) ? (float)y(x) : Y(y(x));
			double xVal = x_.min;
			PointF[] p = new PointF[nPoints];
			for (int n=0; n<nPoints-1; n++)
			{
				p[n] = new PointF(X(xVal), y_x(xVal));
				xVal *= ratio;
			}
			p[nPoints-1] = new PointF(X(x_.max), y_x(x_.max));	// 最後の点

			if (size != 0)
				gr_.Clip = new Region(new RectangleF(
						x_.org, y0, X(x_.max)-X(x_.min), size));

			aPen.LineJoin = LineJoin.Round;
			gr_.SmoothingMode = mode;
			Draw(aPen, p);						// 折れ線または曲線の描画
			gr_.SmoothingMode = SmoothingMode.Default;
			gr_.Clip = new Region();			// 描画領域の限定を解除
		}

		// 全体の目盛線の描画
		private void DrawScale(Values value, Action<float> DrawLine)
		{
			int logMin = (int)Floor(Log10(value.min));
			int loop = (int)Floor(Log10(value.max)) - logMin;

			DrawLine(value.min);	// 最小値に対応する線
			// 座標値の増分
			float du = (float)Pow(10.0, logMin);
			// 最小値に対応する線の次の座標値
			float u1 = (float)(Floor(value.min/du) + 1)*du;

			for (int n=0; n<=loop; n++)
			{
				float uMax = Min(10*du, value.max);
				for (float u=u1; u<uMax*0.99f; u+=du) DrawLine(u);

				du = uMax;			// 値の増分を 10 倍する
				u1 = du;			// 次の ｢for ループ｣ の最初の値
			}
			DrawLine(value.max);	// 最大値に対応する線
		}

		// “Debug”時にのみモードを検査する
		[Conditional("DEBUG")]
		private void Test(Mode mode ,string str)
		{
			if (mode_ != mode) return;
			MessageBox.Show(str);
			Environment.Exit(0);	// アプリケーションの強制終了
		}

		// 10 のべき乗かどうかの検査
		private Func<float, bool> PowersOf10 = x =>
			Abs(Log10(x) - Round(Log10(x))) < 0.01;
	}
}
